% Smooth your Z!
% Jon Barron, 2011. jonbarron@gmail.com. Use at your own risk.
% Modified to run on GNU/Octave by Mathieu Dubois (mdubois@ensta.fr)
function Zsmooth = inpaintZ(Z, lambda_grad)

DEBUG=true;

lambda_curve = 1;
lambda_constrain = 10^3;

addpath(genpath('./bmorph'));

%fig = figure; visualizeZ(Z); drawnow;

% Z(bmorph(isnan(Z), true([3,3]))) = nan;

Zvalid = ~isnan(Z);

if exist('imdilate')
  fidx = find(imdilate(~Zvalid, true([3,3])));
else
  fidx = find(bmorph(~Zvalid, true([3,3])));
end

if DEBUG
  fprintf(2, 'fidx:\n')
  %disp(fidx)
end

[fi1, fj1] = ind2sub(size(Z), fidx);
fi0 = fi1 - 1;
fi2 = fi1 + 1;
fj0 = fj1 - 1;
fj2 = fj1 + 1;

fis = [fi0, fi1, fi2];
fjs = [fj0, fj1, fj2];

keepi = all(fis <= size(Z,1) & fis >= 1,2);
keepj = all(fjs <= size(Z,2) & fjs >= 1,2);

idx1 = [sub2ind(size(Z), fi0(keepi), fj1(keepi)), sub2ind(size(Z), fi1(keepi), fj1(keepi)), sub2ind(size(Z), fi2(keepi), fj1(keepi))];
idx2 = [sub2ind(size(Z), fi1(keepj), fj0(keepj)), sub2ind(size(Z), fi1(keepj), fj1(keepj)), sub2ind(size(Z), fi1(keepj), fj2(keepj))];

idx_curve = [idx1; idx2];
idx_curve = idx_curve(sum(~isnan(Z(idx_curve)),2) < 3,:);

idx_grad = [idx_curve(:,1:2); idx_curve(:,2:3)];
idx_grad = idx_grad(sum(~isnan(Z(idx_grad)),2) < 2,:);

if DEBUG
  fprintf(2, 'idx_curve(%d, %d):\n', size(idx_curve))
  %disp(idx_curve)
  fprintf(2, 'idx_grad(%d, %d):\n', size(idx_grad))
  %disp(idx_grad)
end

if exist('imdilate')
  keep = find(imdilate(~Zvalid, true([5,5])));
else
  keep = find(bmorph(~Zvalid, true([5,5])));
end

if DEBUG
  fprintf(2, 'keep:')
  %disp(keep)
end

Zkeep = Z(keep);

% Octave ismember does not return a matrix of the right shape
% (http://lists.gnu.org/archive/html/octave-bug-tracker/2010-05/msg00083.html)
if exist('OCTAVE_VERSION')
  %fprintf(2, 'Octave specific code\n')
  idx_curve_shape = size(idx_curve);
  [junk, idx_curve] = ismember(idx_curve, keep);
  idx_curve = reshape(idx_curve, idx_curve_shape);
  idx_grad_shape = size(idx_grad);
  [junk, idx_grad] = ismember(idx_grad, keep);
  idx_grad = reshape(idx_grad, idx_grad_shape);
else
  %fprintf(2, 'Matlab specific code\n')
  [junk, idx_curve] = ismember(idx_curve, keep);
  [junk, idx_grad] = ismember(idx_grad, keep);
end

if DEBUG
  fprintf(2, 'idx_curve(%d, %d):\n', size(idx_curve))
  %disp(idx_curve)
  fprintf(2, 'idx_grad(%d, %d):\n', size(idx_grad))
  %disp(idx_grad)
end

Zkeep_valid = ~isnan(Zkeep);

n = size(Zkeep,1);

disp(n)

Acurve = 2*sparse(1:size(idx_curve,1), idx_curve(:,2), 1, size(idx_curve,1), n) - sparse(1:size(idx_curve,1), idx_curve(:,1), 1, size(idx_curve,1), n) - sparse(1:size(idx_curve,1), idx_curve(:,3), 1, size(idx_curve,1), n);
bcurve = sparse(size(Acurve,1),1);

Agrad = sparse(1:size(idx_grad,1), idx_grad(:,1), 1, size(idx_grad,1), n) - sparse(1:size(idx_grad,1), idx_grad(:,2), 1, size(idx_grad,1), n);
bgrad = sparse(size(Agrad,1),1);

Aeq = speye([n, n]);
Aeq = Aeq(Zkeep_valid,:);
beq = Zkeep(Zkeep_valid);

A = [lambda_curve*Acurve; lambda_grad*Agrad; lambda_constrain*Aeq];
b = [lambda_curve*bcurve; lambda_grad*bgrad; lambda_constrain*beq];

if DEBUG
  fprintf(2, 'Construction of A and B\n');
  fprintf(2, 'A(%d, %d)\n', size(A));
  fprintf(2, 'b(%d, %d)\n', size(b));
end

whos("A", "b");
disp(numel(find(A)));

%keyboard
save memory-exhausted A b
tic; X = A \ b; toc;
%tic;
%X = pcg(A, b');
%toc;
%tic;
%X = lsqrSOL(rows(A), columns(A), A, b, 0, 1e-6, 1e-6, 1e8, 1000, 1);
%toc;

Zsmooth = Z;
Zsmooth(keep) = X;

%figure(fig); imagesc([visualizeZ([Z, Zsmooth]); visualizeNormals([Z, Zsmooth])]); imtight;

